/*
 *  Copyright (C) 2015  Malte Humann
 *
 *  This file is part of the Time Series Anomaly Detection (TSAD).
 *
 *  TSAD is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TSAD is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with TSAD.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.tzi.imonitor;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.MathContext;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.jtransforms.fft.DoubleFFT_1D;
import org.rosuda.REngine.REXP;
import org.rosuda.REngine.REXPMismatchException;
import org.rosuda.REngine.REngineException;
import org.rosuda.REngine.Rserve.RConnection;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Analyzes the new performance data entries provided by the
 * {@link SpoolFilesReader} in the context of their corresponding time series
 * for the occurrence of anomalies and sends reports to Icinga accordingly.
 * <p>
 * Also provides {@link #start()} and {@link #stop()} methods to safely stop the
 * execution of this thread.
 */
public class CheckTimeSeries implements Runnable {
	/** Reference to the current thread itself to be able to safely stop it. */
	private volatile Thread cts;
	/**
	 * The SQL database connection prepared and provided by the
	 * {@link AnomalyDetection}.
	 */
	private Connection conn;
	/** Treat values in the [-NZL, +NZL] interval as zero. */
	private static final double NEAR_ZERO_LIMIT = 1e-5;
	/**
	 * The precision used for comparing the target against the upper/lower
	 * bounds.
	 **/
	private static final MathContext BOUNDS_PRECISION = new MathContext(5);
	/**
	 * The number of potential candidates that are chosen from the periodogram
	 * for further refinement.
	 */
	private static final int NUMBER_OF_CANDIDATES = 10;
	/**
	 * The number of time steps that a prediction can be shifted for a best fit.
	 */
	private static final int TIME_SERIES_ALLOWED_LAG = 2;
	private static final int EXIT_FAILURE = 1;
	private static final int EXIT_SUCCESS = 0;
	private static final SimpleDateFormat dateFormat = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss");
	private final boolean send_warnings;
	private final boolean reuseIcingaObjectDefinitions;
	private final String sendNscaConfigPath;
	private final String sendNscaHost;
	/** The time step interval to check for anomalies. */
	private final int window;
	/**
	 * The number of acceptable anomalies in the interval specified in
	 * {@link #window}.
	 */
	private final int threshold;
	/** <code>1 / window</code> */
	private final double percentThresholdWarn;
	/** <code>threshold / window</code> */
	private final double percentThresholdCrit;
	/** The TSAD prefix specified in the tsad.properties file. */
	private final String prefix;
	private final static Logger LOGGER = LoggerFactory
			.getLogger(CheckTimeSeries.class);

	/** Icinga state IDs: 0=OK, 1=WARNING, 2=CRITICAL, 3=UNKNOWN. */
	private static enum STATUS {
		OK, WARN, CRIT, UNKNOWN
	}

	/**
	 * Set up send_nsca and some other Icinga communication related parameters
	 * and prepares the warning and critical thresholds.
	 * 
	 * @param connection
	 *            SQL database connection to the database used by
	 *            {@link SpoolFilesReader}
	 */
	public CheckTimeSeries(Connection connection) {
		conn = connection;
		prefix = Configuration.getProperty("prefix");
		sendNscaConfigPath = Configuration.getProperty("send_nsca_config");
		sendNscaHost = Configuration.getProperty("send_nsca_server");
		window = Integer.parseInt(Configuration.getProperty("window"));
		threshold = Integer.parseInt(Configuration.getProperty("threshold"));
		percentThresholdWarn = 1 / (double) window;
		percentThresholdCrit = threshold / (double) window;
		send_warnings = Boolean.parseBoolean(Configuration
				.getProperty("send_warnings"));
		reuseIcingaObjectDefinitions = Boolean.parseBoolean(Configuration
				.getProperty("reuse_icinga_object_definitions"));
	}

	/** Starts a new thread for the current instance. */
	public void start() {
		cts = new Thread(this);
		cts.start();
	}

	/** Stops the execution of the current {@link #cts thread} safely. */
	public synchronized void stop() {
		LOGGER.debug("Stopping.");
		Thread tmp = cts;
		cts = null;
		tmp.interrupt();
	}

	/**
	 * Waits for notifications from the {@link SpoolFilesReader} that new
	 * performance data entries are available and initiates the processing of
	 * those new entries.
	 */
	@Override
	public void run() {
		// handle remaining time series (after restart for example)
		LOGGER.trace("handle remaining time series");
		handleUnprocessTimeSeries();

		Thread thisThread = Thread.currentThread();
		while (cts == thisThread) {
			// wait for new performance data
			try {
				synchronized (this) {
					wait();
				}
			} catch (InterruptedException e) {
				if (cts == null) {
					LOGGER.debug("Received shut down interrupt.");
				} else {
					LOGGER.warn("Interrupting.", e);
				}
				return;
			}

			// handle new time series
			LOGGER.trace("handle new time series");
			handleUnprocessTimeSeries();
		}
	}

	/**
	 * Initiates the processing of unprocessed time series (that includes new
	 * data as well as old yet unprocessed entries).
	 */
	private void handleUnprocessTimeSeries() {
		Set<Integer> pendingTimeSeries = getPendingTimeSeriesIds();
		for (int i : pendingTimeSeries) {
			checkTimeSeries(i);
		}
	}

	/**
	 * Retrieves the IDs of all time series that have unprocessed entries left.
	 * 
	 * @return a set of time series IDs with yet unprocessed entries
	 */
	private Set<Integer> getPendingTimeSeriesIds() {
		Set<Integer> result = new HashSet<Integer>();

		try (Statement stmt = conn.createStatement()) {
			ResultSet rs = stmt.executeQuery("SELECT id FROM timeseries "
					+ "WHERE lastprocessed < (SELECT MAX(timestamp) FROM "
					+ "perfdata WHERE timeseries_id=id)");
			while (rs.next()) {
				result.add(rs.getInt("id"));
			}
		} catch (SQLException e) {
			LOGGER.warn("Unable to retrieve IDs of unprocessed time series.", e);
		}

		LOGGER.debug("pending time series IDs: " + result);
		return result;
	}

	/**
	 * Checks the pending entries of the time series specified by the given ID
	 * for anomalous entries.
	 * <p>
	 * Uses the last estimated period (if available) to detrend and preclean the
	 * time series. Then predicts and evaluates the current entry based on this
	 * data and stores the result in the database. If reasonable a report will
	 * be sent to Icinga.
	 * <p>
	 * If an error occurs the analysis of the current time series is canceled to
	 * prevent missing intermediate predictions and scores.
	 * 
	 * @param id
	 *            the ID of the time series to be analyzed
	 */
	private void checkTimeSeries(int id) {
		LOGGER.trace("Check time series with ID " + id);
		// retrieve full time series
		TimeSeriesData tsd = selectTimeSeriesData(id);
		if (tsd == null) {
			return;
		}
		int rows = tsd.timestamps.size();
		int lastEstimatedPeriod;
		try {
			lastEstimatedPeriod = selectEstimatedPeriod(id);
		} catch (SQLException e) {
			LOGGER.warn("Unable to select last estimated period.", e);
			return;
		}
		int resumeIndex = rows - countUnprocessedRows(id);
		if (resumeIndex >= rows) {
			LOGGER.debug("No unprocessed entries remaining.");
		}
		for (int i = resumeIndex; i < rows; ++i) {
			// manually convert the Double list to a double array (since toArray
			// only supports the same type, i.e., Double (instead of double) ...
			List<Double> timeseries = tsd.values.subList(0, i + 1);
			int size = timeseries.size();
			double[] data = new double[size];
			for (int j = 0; j < data.length; ++j) {
				data[j] = timeseries.get(j);
			}

			// detrend via STL
			double[] trend;
			try {
				trend = trend(data, lastEstimatedPeriod);
			} catch (REngineException | REXPMismatchException e) {
				LOGGER.error("Unable to use R.", e);
				return;
			}
			double[] detrended = detrend(data, trend);

			// preclean via IQR
			double[] precleaned = preclean(detrended);

			int estimatedPeriod = estimatePeriod(precleaned);

			double prediction;

			// predict latest value; if no period was found, just use the
			// previous value (or if its the first value, just copy that value)
			if (estimatedPeriod == -1) {
				if (i >= 1) {
					prediction = tsd.values.get(i - 1);
				} else {
					prediction = tsd.values.get(i);
				}
			} else {
				// check if detrended data can be reused; update otherwise
				if (estimatedPeriod != lastEstimatedPeriod) {
					try {
						trend = trend(data, estimatedPeriod);
					} catch (REngineException | REXPMismatchException e) {
						LOGGER.error("Unable to use R.", e);
						return;
					}
					detrended = detrend(data, trend);
					lastEstimatedPeriod = estimatedPeriod;
				}
				prediction = predictValue(trend, detrended, estimatedPeriod);
			}

			boolean anomaly = interpretPrediction(tsd.values, tsd.predictions,
					prediction, i);
			// update locally cached data
			tsd.predictions.set(i, prediction);
			tsd.scores.set(i, anomaly);

			LOGGER.debug(String.format("ID %d (%d):\t%d\t%d\t%f, %f, %b", id,
					i, estimatedPeriod, tsd.timestamps.get(i),
					tsd.values.get(i), prediction, anomaly));

			double score = report(i, tsd.scores);
			STATUS status = getStatusFor(score);

			// report status to Icinga
			if (!reuseIcingaObjectDefinitions
					|| (status == STATUS.CRIT || (status == STATUS.WARN && send_warnings))) {
				// prepare send_nsca content

				// hostname
				StringBuilder message = new StringBuilder(tsd.hostname);
				message.append('+');

				// optional service description if it's a service
				if (reuseIcingaObjectDefinitions) {
					if (!tsd.servicedescription.equals("")) {
						message.append(tsd.servicedescription).append('+');
					}
				} else {
					message.append(prefix);
					message.append(tsd.servicedescription);
					message.append('-').append(tsd.perfdatalabel);
					message.append('+');
				}

				// status
				if (reuseIcingaObjectDefinitions) {
					message.append(tsd.states.get(i));
				} else {
					message.append(status.ordinal());
				}

				// custom json message (used by iMonitor)
				message.append("+{\"type\":\"TSAD\",");
				// timestamp of the original report; timestamp of the TSAD
				// report is provided through send_nsca
				message.append("\"timestamp\":\"");
				message.append(dateFormat.format(new Date(
						tsd.timestamps.get(i) * 1000l)));
				message.append("\",");
				// status
				message.append("\"status\":").append(status.ordinal())
						.append(',');
				// message
				message.append(String.format(
						"\"message\":\"%s %s - %.0f%% anomalous\"}",
						tsd.perfdatalabel, status, score * 100));

				// performance data
				message.append(String.format("|'%s%s'=%.2f;%.2f;%.2f;", prefix,
						tsd.perfdatalabel.replace("'", "''"), score,
						percentThresholdWarn, percentThresholdCrit));
				message.append('\n');

				int sendNscaStatus = sendNsca(message.toString());
				if (sendNscaStatus != EXIT_SUCCESS) {
					// warning already logged through sendNsca(String)
					return;
				}
			}

			// update prediction, last processed timestamp and estimated period
			updatePrediction(tsd.timestamps.get(i), id, prediction, anomaly);
			updateMetaData(tsd.timestamps.get(i), estimatedPeriod, id);
			// TODO truncate data if necessary (e.g., store only 90 days)
		}
	}

	/**
	 * Counts the number of anomalies in the latest {@link #window x} entries.
	 * 
	 * @param current
	 *            the position of the current evaluated entry in the time series
	 * @param scores
	 *            the list of all scores for the corresponding time series so
	 *            far
	 * @return
	 */
	private double report(int current, List<Boolean> scores) {
		int anomalies = 0;
		for (int i = current, limit = Math.max(0, current - window); i > limit; --i) {
			if (scores.get(i)) {
				++anomalies;
			}
		}
		return anomalies / (double) window;
	}

	/**
	 * Converts the given score to a Icinga status.
	 * 
	 * @param score
	 *            the score to be converted
	 * @return the corresponding status
	 */
	private STATUS getStatusFor(double score) {
		if (score >= percentThresholdCrit) {
			return STATUS.CRIT;
		} else if (score >= percentThresholdWarn) {
			return STATUS.WARN;
		} else {
			return STATUS.OK;
		}
	}

	/**
	 * Sends the given message to Icinga invoking send_nsca via an external bash
	 * command.
	 * 
	 * @param message
	 *            the messages passed through send_nsca
	 * @return {@link #EXIT_SUCCESS} if send_nsca returned without error,
	 *         {@link #EXIT_FAILURE} otherwise
	 */
	private int sendNsca(String message) {
		// prepare send_nsca command
		String checkResult = String.format("printf \"%s\"",
				message.replace("\"", "\\\"").replace("%", "%%"));
		String sendNscaCommand = String.format(" | send_nsca -H %s -d + -c %s",
				sendNscaHost, sendNscaConfigPath);
		LOGGER.debug(checkResult + sendNscaCommand);
		ProcessBuilder sendNscaPB = new ProcessBuilder("/bin/bash", "-c",
				checkResult + sendNscaCommand);
		Process sendNsca;
		try {
			sendNsca = sendNscaPB.start();
		} catch (IOException e) {
			LOGGER.warn("Unable to inform Icinga through send_nsca.", e);
			return EXIT_FAILURE;
		}

		try {
			int status = sendNsca.waitFor();
			if (status == EXIT_SUCCESS) {
				LOGGER.trace("send_nsca finished with exit status " + status);
			} else {
				LOGGER.debug("send_nsca finished with exit status " + status);
			}
			return status;
		} catch (InterruptedException e) {
			if (cts != null) {
				LOGGER.warn(
						"Unknown if send_nsca command finished (interrupted).",
						e);
			}
			return EXIT_FAILURE;
		}
	}

	/**
	 * Calculates upper and lower bounds based on the prediction and compares
	 * this interval against the observed value.
	 * 
	 * @param values
	 *            the original time series measurements received from Icinga
	 * @param predictions
	 *            the predicted time series values so far
	 * @param prediction
	 *            the prediction for the current entry
	 * @param currentIndex
	 *            the index of the current entry
	 * @return true of the observed value lies outside the interval (and
	 *         therefore is considered an anomaly), false otherwise
	 */
	private boolean interpretPrediction(List<Double> values,
			List<Double> predictions, double prediction, int currentIndex) {
		if (currentIndex <= 1) {
			return false;
		}
		// omit latest value/prediction/error
		double[] errors = new double[currentIndex - 1];
		for (int i = 0; i < errors.length; i++) {
			errors[i] = Math.abs(values.get(i) - predictions.get(i));
		}
		Arrays.sort(errors);
		InterquartileRange iqr = new InterquartileRange(errors);
		double median = quantile(errors, .5);

		BigDecimal upper = new BigDecimal(prediction
				+ Math.abs(iqr.thirdQuartile - median) + 1.5 * iqr.iqr,
				BOUNDS_PRECISION);
		BigDecimal lower = new BigDecimal(prediction
				- Math.abs(median - iqr.firstQuartile) - 1.5 * iqr.iqr,
				BOUNDS_PRECISION);

		BigDecimal target = new BigDecimal(values.get(currentIndex),
				BOUNDS_PRECISION);
		if (target.compareTo(upper) > 0 || target.compareTo(lower) < 0) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Predicts the current value of the time series based on the period length
	 * and the detrended data. A short lag for the position in the period is
	 * allowed as speficied by {@link #TIME_SERIES_ALLOWED_LAG}.
	 * <p>
	 * Since the predicted value doesn't recognize the trend a final step is
	 * needed that adds the trend again to complete the prediction.
	 * 
	 * @param trend
	 *            the trend of the time series up to and including the current
	 *            entry
	 * @param detrended
	 *            the time series with the trend removed
	 * @param period
	 *            length of the period
	 * @return the predicted value for the current time series
	 */
	private double predictValue(double[] trend, double[] detrended, int period) {
		// determine figure using the median (not mean)
		double[] figure = new double[period];
		for (int i = 0; i < period; ++i) {
			double[] data = new double[(detrended.length - i) / period];
			for (int j = 0; j < data.length; ++j) {
				data[j] = detrended[i + j * period];
			}
			Arrays.sort(data);
			figure[i] = quantile(data, 0.5);
		}

		// allow of a slightly lagged index
		int index = (detrended.length - 1) % period;
		// 'lowest' index (inclusive)
		int left = (index - TIME_SERIES_ALLOWED_LAG) % period;
		if (left < 0) {
			left = left + period;
		}
		// 'highest' index (exclusive)
		int right = (index + TIME_SERIES_ALLOWED_LAG + 1) % period;

		// latest measurement (that should be predicted) w/o trend
		double target = detrended[detrended.length - 1];

		double prediction = figure[left];
		double minError = Math.abs(prediction - target);

		for (int i = ++left % period; i != right; i = ++i % period) {
			double error = Math.abs(figure[i] - target);
			if (error < minError) {
				minError = error;
				prediction = figure[i];
			}
		}

		// add trend again
		return prediction + trend[trend.length - 1];
	}

	/**
	 * Estimates the period length based on the periodogram and autocorrelation
	 * of the time series.
	 * 
	 * @param data
	 *            the original measurements received from Icinga
	 * @return the estimated length of the period or -1 if the length couldn't
	 *         be determined
	 */
	private int estimatePeriod(double[] data) {
		// the first half needs at least a third value as proper period
		// candidate
		if (data.length / 2 <= 3) {
			return -1;
		}

		DoubleFFT_1D dfft = new DoubleFFT_1D(data.length);

		// apply fast Fourier transformation (FFT)
		double[] fft = new double[2 * data.length];
		System.arraycopy(data, 0, fft, 0, data.length);
		dfft.realForwardFull(fft);

		// create the periodogram based on the FFT result
		final double[] periodogram = new double[data.length];
		for (int i = 0; i < periodogram.length; ++i) {
			periodogram[i] = Math.pow(fft[2 * i], 2)
					+ Math.pow(fft[2 * i + 1], 2);
		}

		// apply inverse FFT on the periodogram to receive the autocorrelation
		// function results
		double[] ifft = new double[2 * data.length];
		System.arraycopy(periodogram, 0, ifft, 0, data.length);
		dfft.realInverseFull(ifft, true);
		double[] acf = new double[data.length];
		for (int i = 0; i < acf.length; ++i) {
			acf[i] = ifft[2 * i]; // take only the real part
		}

		// get top NUMBER_OF_CANDIDATES period candidates
		// only use the first half of the periodogram (Nyquist)
		Integer[] periodCandidates = new Integer[data.length / 2];
		for (int i = 0; i < periodCandidates.length; ++i) {
			periodCandidates[i] = i;
		}

		// skip indices 0 and 1, since 0 represents a constant, not a
		// sine/cosine, and 1 represents a period as long as the whole
		// series
		Arrays.sort(periodCandidates, 2, periodCandidates.length - 1,
				new Comparator<Integer>() {
					@Override
					public int compare(Integer i1, Integer i2) {
						return Double.compare(periodogram[i1], periodogram[i2]);
					}
				});
		;
		// after sorting, select the top NUMBER_OF_PERIODS periods / periodogram
		// indices
		periodCandidates = Arrays.copyOfRange(periodCandidates, Math.max(
				periodCandidates.length - (NUMBER_OF_CANDIDATES + 1), 2),
				periodCandidates.length - 1);

		// refine candidates and select the 'best' (highest ACF value)
		Set<Integer> refinedCandidates = refineCandidates(acf, periodCandidates);
		return selectCandidate(acf, refinedCandidates);
	}

	/**
	 * Selects the candidate with the highest autocorrelation value. If multiple
	 * candidates are tied the larger period length is used as deciding factor.
	 * 
	 * @param acf
	 *            the autocorrelation values of the time series
	 * @param refinedCandidates
	 *            the candidate set
	 * @return the best candidate from the given set
	 */
	private int selectCandidate(double[] acf, Set<Integer> refinedCandidates) {
		double highestAcf = Double.NEGATIVE_INFINITY;
		int period = -1;
		for (Integer i : refinedCandidates) {
			double acfValue = acf[i];
			if (acfValue > highestAcf || (acfValue == highestAcf && i > period)) {
				highestAcf = acfValue;
				period = i;
			}
		}

		return period;
	}

	/**
	 * Refines candidates chosen from the periodogram using the autocorrelation
	 * values of those candidates.
	 * <p>
	 * Adjacent candidates are combined into one additional candidate.
	 * <p>
	 * Candidates that are not considered local maxima in the autocorrelation
	 * are discarded.
	 * 
	 * @param acf
	 *            the autocorrelation values of the time series
	 * @param periodCandidates
	 *            potential candidates retrieved from the periodogram
	 * @return a set of refined candidates
	 */
	private Set<Integer> refineCandidates(double[] acf,
			Integer[] periodCandidates) {
		// select only valid candidates, i.e., peaks in the ACF
		Set<Integer> refinedCandidates = new HashSet<Integer>();
		for (int i = 0; i < periodCandidates.length; ++i) {
			int period = validateCandidate(acf, periodCandidates[i]);
			if (period != -1) {
				refinedCandidates.add(period);
			}
		}
		// combine adjacent candidates to one interval
		Integer[] searchableCandidates = periodCandidates.clone();
		Arrays.sort(searchableCandidates);
		for (int i = 0; i < periodCandidates.length; ++i) {
			if (Arrays.binarySearch(searchableCandidates,
					periodCandidates[i] + 1) >= 0) {
				int period = validateExpandedCandidate(acf, periodCandidates[i]);
				if (period != -1) {
					refinedCandidates.add(period);
				}
			}
		}

		return refinedCandidates;
	}

	private int validateExpandedCandidate(double[] acf, int k) {
		return validateCandidate(acf, k, 1);
	}

	private int validateCandidate(double[] acf, int k) {
		return validateCandidate(acf, k, 0);
	}

	/**
	 * Checks if autocorrelation interval around a given candidate contains a
	 * local maximum surrounded by a upward-downward trend.
	 * 
	 * @param acf
	 *            the autocorrelation values of the time series
	 * @param k
	 *            the position of the candidate to validate
	 * @param range
	 *            the range of adjacent candidates to also consider
	 * @return a refined candidate index if valid or -1 otherwise
	 */
	private int validateCandidate(double[] acf, int k, int range) {
		int n = acf.length;
		int lower = (int) Math
				.floor((n / (k + range + 1) + n / (k + range)) / 2 - 1);
		int upper = (int) Math.ceil(((n / k) + (n / (k - 1))) / 2 + 1);
		// limit interval to [2, n/2]
		lower = Math.max(2, lower);
		upper = Math.min(n / 2, upper);

		boolean upwardSlope = false;
		int candidateIndex = -1;
		double max = Double.NEGATIVE_INFINITY;
		int index = -1;

		for (int i = lower + 1; i < upper; ++i) {
			double slope = acf[i] - acf[i - 1];
			if (slope > NEAR_ZERO_LIMIT) {
				upwardSlope = true;
				candidateIndex = i;
			} else if (slope < -NEAR_ZERO_LIMIT) {
				// use larger period if both are valid
				if (upwardSlope && acf[candidateIndex] >= max) {
					max = acf[candidateIndex];
					index = candidateIndex;
				}
				upwardSlope = false;
			} else { // slope is (about) zero
				// use larger period if the current point lies on a plateau
				if (upwardSlope) {
					candidateIndex = i;
				}
			}
		}

		return index;
	}

	/**
	 * Trims outliers of the given time series to 1.5 times the interquartile
	 * range.
	 * 
	 * @param data
	 *            the unmodified time series entries as received by Icinga
	 * @return a new copy of the time series with trimmed outliers
	 */
	private double[] preclean(double[] data) {
		if (data.length < 2) {
			return data.clone();
		}

		InterquartileRange iqr = new InterquartileRange(data);

		if (iqr.nearZero) {
			// don't discard the data by trimming it all to about zero
			return data.clone();
		}

		double[] cleaned = new double[data.length];
		for (int i = 0; i < data.length; ++i) {
			if (data[i] > iqr.upper) {
				cleaned[i] = iqr.upper;
			} else if (data[i] < iqr.lower) {
				cleaned[i] = iqr.lower;
			} else {
				cleaned[i] = data[i];
			}
		}
		return cleaned;
	}

	/**
	 * Determines the trend of the given time series using STL via R.
	 * 
	 * @param data
	 *            the time series values
	 * @param period
	 *            the (current estimate of the) period length
	 * @return the trend or a list of zeros
	 * @throws REngineException
	 * @throws REXPMismatchException
	 */
	private double[] trend(double[] data, int period) throws REngineException,
			REXPMismatchException {
		if (period < 2 || data.length <= 2 * period) {
			return new double[data.length];
		}

		// get trend via STL
		double[] result;
		String request = String.format("stl(ts(x, frequency=%d), "
				+ "s.window=\"periodic\", robust=TRUE)", period);
		RConnection connection = new RConnection();
		connection.assign("x", data);
		REXP r = connection.eval(request);
		result = r.asList().at(0).asDoubles();

		int size = data.length;
		return Arrays.copyOfRange(result, size, size + size);
	}

	/**
	 * Removes the additive trend from the given time series.
	 * 
	 * @param data
	 *            the time series
	 * @param trend
	 *            the trend
	 * @return a new copy of the time series without the trend
	 */
	private double[] detrend(double[] data, double[] trend) {
		double[] detrended = new double[data.length];
		for (int i = 0; i < data.length; ++i) {
			detrended[i] = data[i] - trend[i];
		}
		return detrended;
	}

	/**
	 * Inquires the count of unprocessed performance data entries for the given
	 * time series from the database.
	 * 
	 * @param id
	 *            a time series ID
	 * @return the number of unprocessed entries for the given time series
	 */
	private int countUnprocessedRows(int id) {
		int rows = -1;
		try (PreparedStatement pstmt = conn.prepareStatement("SELECT COUNT(*) "
				+ "FROM perfdata WHERE timestamp > (SELECT lastprocessed FROM "
				+ "timeseries WHERE id=?) AND timeseries_id=?")) {
			pstmt.setInt(1, id);
			pstmt.setInt(2, id);
			ResultSet rs = pstmt.executeQuery();
			if (rs.next()) {
				rows = rs.getInt(1);
			}
		} catch (SQLException e) {
			LOGGER.warn("Unable to count unprocessed entries.", e);
		}

		return rows;
	}

	/**
	 * Retrieves the latest estimated period length for a given time series from
	 * the database.
	 * 
	 * @param id
	 *            a time series ID
	 * @return the latest known estimated period length of the time series or -1
	 * @throws SQLException
	 */
	private int selectEstimatedPeriod(int id) throws SQLException {
		int period = -1;

		try (PreparedStatement pstmt = conn.prepareStatement("SELECT "
				+ "estimated_period FROM timeseries WHERE id=?")) {
			pstmt.setInt(1, id);
			ResultSet rs = pstmt.executeQuery();
			if (rs.next()) {
				period = rs.getInt(1);
			}
		} catch (SQLException e) {
			// throw to make use of the implicit finally for closing pstmt
			throw e;
		}

		return period;
	}

	/**
	 * Makes a local copy of most of the time series related data from the
	 * database, storing them in a {@link TimeSeriesData} object for simplified
	 * handling.
	 * 
	 * @param id
	 *            a time series ID
	 * @return a local copy of the time series
	 */
	private TimeSeriesData selectTimeSeriesData(int id) {
		TimeSeriesData tsd;
		try (PreparedStatement pstmt = conn.prepareStatement("SELECT "
				+ "hostname, servicedescription, perfdatalabel FROM "
				+ "hostnames h JOIN timeseries t ON h.id = t.hostname_id "
				+ "JOIN servicedescriptions s "
				+ "ON s.id = t.servicedescription_id "
				+ "JOIN perfdatalabels p ON p.id = t.perfdatalabel_id "
				+ "WHERE t.id=?")) {
			pstmt.setInt(1, id);
			ResultSet rs = pstmt.executeQuery();
			if (rs.next()) {
				tsd = new TimeSeriesData(rs.getString(1), rs.getString(2),
						rs.getString(3));
			} else {
				LOGGER.error("Unable to select current time series data.");
				return null;
			}
		} catch (SQLException e) {
			LOGGER.error("Unable to select current time series data.", e);
			return null;
		}

		try (PreparedStatement pstmt = conn.prepareStatement("SELECT "
				+ "timestamp, value, predicted, score, state FROM perfdata"
				+ " WHERE timeseries_id=? ORDER BY timestamp ASC")) {
			pstmt.setInt(1, id);
			ResultSet rs = pstmt.executeQuery();
			while (rs.next()) {
				tsd.timestamps.add(rs.getInt(1));
				tsd.values.add(rs.getDouble(2));
				tsd.predictions.add(rs.getDouble(3));
				tsd.scores.add(rs.getBoolean(4));
				tsd.states.add(rs.getInt(5));
			}
		} catch (SQLException e) {
			LOGGER.error("Unable to select current time series data.", e);
			return null;
		}

		return tsd;
	}

	/**
	 * Updates the meta data entry of a time series in the database.
	 * 
	 * @param timestamp
	 *            timestamp of the latest processed performance data entry
	 * @param estimatedPeriod
	 *            the latest estimated period for the time series
	 * @param id
	 *            a time series ID
	 */
	private void updateMetaData(int timestamp, int estimatedPeriod, int id) {
		try (PreparedStatement pstmt = conn.prepareStatement("UPDATE "
				+ "timeseries SET lastprocessed=?, estimated_period=? WHERE "
				+ "id=?")) {
			pstmt.setInt(1, timestamp);
			pstmt.setInt(2, estimatedPeriod);
			pstmt.setInt(3, id);
			pstmt.executeUpdate();
		} catch (SQLException e) {
			if (cts != null) {
				LOGGER.warn("Unable to update meta data.", e);
			}
		}
	}

	/**
	 * Stores the prediction and anomaly score for the given time series in the
	 * database.
	 * 
	 * @param timestamp
	 *            timestamp of the latest processed performance data entry
	 * @param id
	 *            a time series ID
	 * @param prediction
	 *            the predicted value of the time series for the specified
	 *            timestamp
	 * @param score
	 *            a score stating if the value at this timestamp is anomalous
	 *            (true) or normal (false)
	 */
	private void updatePrediction(int timestamp, int id, double prediction,
			boolean score) {
		// update database
		try (PreparedStatement pstmt = conn.prepareStatement("UPDATE perfdata "
				+ "SET predicted=?, score=? WHERE timestamp=? AND "
				+ "timeseries_id=?")) {
			pstmt.setDouble(1, prediction);
			pstmt.setBoolean(2, score);
			pstmt.setInt(3, timestamp);
			pstmt.setInt(4, id);
			pstmt.executeUpdate();
		} catch (SQLException e) {
			if (cts != null) {
				LOGGER.warn("Unable to update time series data.", e);
			}
		}
	}

	/**
	 * Finds the quantile according to the given percentage.
	 * <p>
	 * Assumes that the data is sorted ascending!
	 * 
	 * @param data
	 *            ascending sorted data
	 * @param percent
	 *            the percentile specifing the desired quantile
	 * @return the quantile
	 */
	private double quantile(double[] data, double percent) {
		double quartile;
		double index = data.length * percent;
		int i = (int) index;

		if (index % 1 == 0) {
			quartile = (data[i - 1] + data[i]) / 2.0;
		} else {
			quartile = data[i];
		}

		return quartile;
	}

	/**
	 * Contains various measures resulting from the interquartile range
	 * calculation for a given time series.
	 */
	private class InterquartileRange {
		private final double firstQuartile;
		private final double thirdQuartile;
		private final double upper;
		private final double lower;
		private final double iqr;
		private final boolean nearZero;

		public InterquartileRange(double[] data) {
			double[] sorted = data.clone();
			Arrays.sort(sorted);

			firstQuartile = quantile(sorted, .25);
			thirdQuartile = quantile(sorted, .75);

			iqr = thirdQuartile - firstQuartile;
			lower = firstQuartile - 1.5 * iqr;
			upper = thirdQuartile + 1.5 * iqr;

			nearZero = (lower == upper || Math.abs(iqr) < NEAR_ZERO_LIMIT);
		}
	}

	/**
	 * Contains various time series related data for simplified access.
	 */
	private class TimeSeriesData {
		// TODO get row count beforehand to initialize the array properly?
		private final List<Integer> timestamps = new ArrayList<>();
		private final List<Double> values = new ArrayList<>();
		private final List<Double> predictions = new ArrayList<>();
		private final List<Boolean> scores = new ArrayList<>();
		private final List<Integer> states = new ArrayList<>();
		private final String hostname;
		private final String servicedescription;
		private final String perfdatalabel;

		public TimeSeriesData(String hostname, String servicedescription,
				String perfdatalabel) {
			this.hostname = hostname;
			this.servicedescription = servicedescription;
			this.perfdatalabel = perfdatalabel;
		}
	}
}

