/*
 *  Copyright (C) 2015  Malte Humann
 *
 *  This file is part of the Time Series Anomaly Detection (TSAD).
 *
 *  TSAD is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TSAD is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with TSAD.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.tzi.imonitor;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Reads the files found by the {@link SpoolFilesWatchDog} into the database
 * provided by {@link AnomalyDetection} and informs the corresponding
 * {@link CheckTimeSeries} instance that there is new data to be processed
 * afterwards.
 * <p>
 * Also provides {@link #start()} and {@link #stop()} methods to safely stop the
 * execution of this thread.
 */
public class SpoolFilesReader implements Runnable {
	/** Reference to the current thread itself to be able to safely stop it. */
	private volatile Thread reader;
	/**
	 * Reference to the check time series instance that gets informed after
	 * successful storing the file's content in the database.
	 */
	private CheckTimeSeries checkTimeSeries;
	/**
	 * The SQL database connection prepared and provided by the
	 * {@link AnomalyDetection}.
	 */
	private Connection conn;
	// TODO max capacity?
	/**
	 * The queue used for communication between this reader and the
	 * corresponding {@link SpoolFilesWatchDog}.
	 */
	private BlockingQueue<Path> queue = new LinkedBlockingQueue<>();
	// allowing any number of escaped ' pairs, i.e. '', and one final '= part
	/**
	 * Regular expression that determines the end of a label part of a
	 * performance data entry.
	 */
	private Pattern labelEndPattern = Pattern.compile("[^']('')*'=");
	/**
	 * Regular expression that determines the digits component of a value part
	 * of a performance data entry.
	 */
	private Pattern valuePattern = Pattern.compile("[-0-9.]+");
	private Map<String, Integer> hostnamesCache = new HashMap<String, Integer>();
	private Map<String, Integer> serviceDescriptionCache = new HashMap<String, Integer>();
	private Map<String, Integer> perfDataLabelCache = new HashMap<String, Integer>();
	private Map<String, Integer> timeseriesCache = new HashMap<String, Integer>();
	private final static Logger LOGGER = LoggerFactory
			.getLogger(SpoolFilesReader.class);
	/** The TSAD prefix specified in the tsad.properties file. */
	private final String prefix;

	public SpoolFilesReader(CheckTimeSeries checkTimeSeries,
			Connection connection) {
		this.checkTimeSeries = checkTimeSeries;
		conn = connection;
		prefix = Configuration.getProperty("prefix");
	}

	/**
	 * Adds a given path to a file to this readers internal processing queue.
	 * 
	 * @param file
	 *            path to the file that should be read
	 */
	public void queueFile(Path file) {
		queue.add(file);
	};

	/** Starts a new thread for the current instance. */
	public void start() {
		reader = new Thread(this);
		reader.start();
	}

	/** Stops the execution of the current {@link #reader thread} safely. */
	public synchronized void stop() {
		LOGGER.debug("Stopping.");
		Thread tmp = reader;
		reader = null;
		tmp.interrupt();
	}

	/**
	 * Reads the file at the given path, parses it's content via
	 * {@link #parse(String)} and writes the result into the database via
	 * {@link #insertPerfData(int, int, double, int)}. After successful storing
	 * the performance data in the database the original file will be deleted.
	 * In case something went wrong the file will be kept.
	 * 
	 * @param path
	 *            path to the performance data file to be processed
	 * @return true if all performance data entries from the given file were
	 *         successfully stored in the database, false otherwise
	 */
	private boolean read(Path path) {
		LOGGER.debug("reading " + path);

		// read data from file
		List<String> rawdata = new LinkedList<>();
		Charset charset = Charset.forName("UTF-8");
		try (BufferedReader reader = Files.newBufferedReader(path, charset)) {
			String line = null;
			while ((line = reader.readLine()) != null) {
				// skip blank lines
				if (line.trim().length() > 0) {
					rawdata.add(line);
				}
			}
		} catch (IOException e) {
			LOGGER.warn("Unable to read " + path, e);
			return false;
		}

		// parse data / prepare queries
		List<PerfDataEntry> queries = new LinkedList<>();
		for (String data : rawdata) {
			queries.addAll(parse(data));
			LOGGER.trace("Added '" + data + "'");
		}

		LOGGER.debug("read " + queries.size() + " entries");

		// write data into database
		int successful = 0;
		for (PerfDataEntry c : queries) {
			// skip own reports
			if (c.perfdatalabel.startsWith(prefix)) {
				LOGGER.debug("skip own reports: " + c.toString());
				continue;
			}
			// get cached key (create missing entries along the way)
			int hostId = getCachedHostnameId(c.hostname);
			int serviceId = getCachedServiceDescriptionId(c.servicedescription);
			int perfdataId = getCachedPerfDataLabelId(c.perfdatalabel);
			if (hostId == -1 || serviceId == -1 || perfdataId == -1) {
				// unable to select or insert corresponding IDs
				LOGGER.debug("unable to select host/service/perfdata ID: "
						+ c.toString());
				continue;
			}
			int tsId = getCachedTimeSeriesId(hostId, serviceId, perfdataId);
			if (tsId == -1) {
				// unable to select or insert corresponding ID
				LOGGER.debug("unable to select time series ID: " + c.toString());
				continue;
			}
			if (insertPerfData(tsId, c.timestamp, c.value, c.state)) {
				LOGGER.debug("succesful inserted (" + tsId + "): "
						+ c.toString());
				++successful;
			} else {
				LOGGER.debug("couldn't insert (" + tsId + "): " + c.toString());
			}
		}

		// if some entries weren't inserted into the database, don't delete the
		// corresponding file, but handle available entries
		if (successful == 0) {
			return false;
		} else if (successful < queries.size()) {
			return true;
		} else {
			// remove successful processed file
			try {
				Files.delete(path);
			} catch (IOException e) {
				LOGGER.warn("Unable to delete '" + path + "' after processing.");
				return true;
			}

			return true;
		}
	}

	/**
	 * Inserts the provided time series data in the corresponding database
	 * field.
	 * 
	 * @param tsId
	 *            time series ID
	 * @param timestamp
	 *            timestamp of the performance data entry
	 * @param value
	 *            value of the performance data entry
	 * @param state
	 *            state reported for the associated host/service
	 * @return true if the data was successfully inserted into the database
	 */
	private boolean insertPerfData(int tsId, int timestamp, double value,
			int state) {
		// check if data already exists to prevent duplicate entries
		try (PreparedStatement pstmt = conn.prepareStatement("SELECT "
				+ "* FROM perfdata WHERE timeseries_id=? AND timestamp=? "
				+ "AND value=?")) {
			pstmt.setInt(1, tsId);
			pstmt.setInt(2, timestamp);
			pstmt.setDouble(3, value);
			ResultSet rs = pstmt.executeQuery();
			if (rs.next()) {
				LOGGER.debug("Unable to insert current perfdata; already existing.");
				return false; // entry already exists
			}
		} catch (SQLException e) {
			LOGGER.warn("Unable to check if current performance data is "
					+ "already stored in the database", e);
			return false;
		}

		// leave predicted empty, since this value isn't known yet
		try (PreparedStatement pstmt = conn.prepareStatement("INSERT "
				+ "INTO perfdata (timeseries_id, timestamp, value, state) "
				+ " VALUES (?, ?, ?, ?)")) {
			pstmt.setInt(1, tsId);
			pstmt.setInt(2, timestamp);
			pstmt.setDouble(3, value);
			pstmt.setInt(4, state);
			pstmt.executeUpdate();
		} catch (SQLException e) {
			LOGGER.warn("Unable to insert current performance data into the "
					+ "database.", e);
			return false;
		}

		return true;
	}

	private int getCachedHostnameId(String key) {
		String table = "hostnames";
		String row = "hostname";
		return getCachedId(hostnamesCache, table, row, key);
	}

	private int getCachedServiceDescriptionId(String key) {
		String table = "servicedescriptions";
		String row = "servicedescription";
		return getCachedId(serviceDescriptionCache, table, row, key);
	}

	private int getCachedPerfDataLabelId(String key) {
		String table = "perfdatalabels";
		String row = "perfdatalabel";
		return getCachedId(perfDataLabelCache, table, row, key);
	}

	private int getCachedId(Map<String, Integer> cache, String table,
			String row, String key) {
		if (cache.containsKey(key)) {
			return cache.get(key);
		} else {
			int id = selectId(table, row, key);
			if (id == -1) {
				// not in database -> insert -> select
				try (PreparedStatement pstmt = conn.prepareStatement(String
						.format("INSERT INTO %s (%s) VALUES(?)", table, row))) {
					pstmt.setString(1, key);
					pstmt.executeUpdate();
				} catch (SQLException e) {
					LOGGER.warn("Unable to insert missing ID into database.", e);
				}
				id = selectId(table, row, key);
			}

			// cache id
			cache.put(key, id);
			return id;
		}
	}

	private int selectId(String table, String row, String key) {
		int id = -1;
		try (PreparedStatement pstmt = conn.prepareStatement(String.format(
				"SELECT id FROM %s WHERE %s=?", table, row))) {
			pstmt.setString(1, key);
			ResultSet rs = pstmt.executeQuery();
			if (rs.next()) {
				// already in database
				id = rs.getInt("id");
			} else {
				// not in database -> insert
				id = -1;
			}
		} catch (SQLException e) {
			LOGGER.warn("Unable to select ID from database.", e);
		}
		return id;
	}

	private int getCachedTimeSeriesId(int hostId, int serviceId, int perfdataId) {
		String key = String.format("%d+%d+%d", hostId, serviceId, perfdataId);

		if (timeseriesCache.containsKey(key)) {
			return timeseriesCache.get(key);
		} else {
			int id = selecTimeSeriestId(hostId, serviceId, perfdataId);
			if (id == -1) {
				// not in database -> insert -> select
				try (PreparedStatement pstmt = conn.prepareStatement("INSERT "
						+ "INTO timeseries (hostname_id, "
						+ "servicedescription_id, perfdatalabel_id, "
						+ "lastprocessed) VALUES (?, ?, ?, ?)")) {
					pstmt.setInt(1, hostId);
					pstmt.setInt(2, serviceId);
					pstmt.setInt(3, perfdataId);
					pstmt.setInt(4, -1);
					pstmt.executeUpdate();
				} catch (SQLException e) {
					LOGGER.warn("Unable to insert missing ID into database.", e);
				}
				id = selecTimeSeriestId(hostId, serviceId, perfdataId);
			}

			// cache id
			timeseriesCache.put(key, id);
			return id;
		}
	}

	private int selecTimeSeriestId(int hostId, int serviceId, int perfdataId) {
		int id = -1;
		try (PreparedStatement pstmt = conn.prepareStatement("SELECT id FROM "
				+ "timeseries WHERE hostname_id=? AND servicedescription_id=? "
				+ "AND perfdatalabel_id=?")) {
			pstmt.setInt(1, hostId);
			pstmt.setInt(2, serviceId);
			pstmt.setInt(3, perfdataId);
			ResultSet rs = pstmt.executeQuery();
			if (rs.next()) {
				// already in database
				id = rs.getInt("id");
			} else {
				// not in database -> insert
				id = -1;
			}
		} catch (SQLException e) {
			LOGGER.warn("Unable to select ID from database.", e);
		}
		return id;
	}

	/**
	 * Converts the given performance data entry into one ore more
	 * {@link PerfDataEntry}s. If the performance data contains multiple label
	 * value pairs a new entry is created for each of those pairs.
	 * 
	 * @param data
	 *            a performance data entry
	 * @return a collection of all performance data entries contained in the
	 *         provided raw entry
	 */
	private Collection<PerfDataEntry> parse(String data) {
		Collection<PerfDataEntry> result = new LinkedList<>();
		String[] entries = data.split("\t");

		// get performance data meta informations
		int index = 1;
		int timestamp = Integer.parseInt(entries[index++].split("::")[1]);
		String hostname = entries[index++].split("::")[1];
		String servicedescription = "";
		if (entries[0].contains("SERVICE")) {
			servicedescription = entries[index++].split("::")[1];
		}

		int state = Integer.parseInt(entries[index + 2].split("::")[1]);

		// get individual performance data
		String perfdata = entries[index]
				.substring(entries[index].indexOf("::") + 2);
		for (int i = 0; i > -1;) {
			int labelstart = i;
			int labelend;
			int valuestart;
			// labels may contain any character according to the documentation
			// so check if the label starts with an ' and therefore should end
			// by '=
			if (perfdata.charAt(i) == '\'') {
				labelstart = i + 1;
				Matcher m = labelEndPattern.matcher(perfdata
						.substring(labelstart + 1));
				m.find();
				// don't check if pattern exist, since it has to
				labelend = m.end();
				valuestart = labelend + 2;
			} else {
				labelend = perfdata.indexOf('=', labelstart + 1);
				valuestart = labelend + 1;
			}
			String label = perfdata.substring(labelstart, labelend);
			int valueend = perfdata.indexOf(';', valuestart);
			if (valueend == -1) {
				valueend = perfdata.indexOf(' ', valuestart);
				if (valueend == -1) {
					valueend = perfdata.length() - 1;
				}

			}

			String valueStr = perfdata.substring(valuestart, valueend);
			Matcher m = valuePattern.matcher(valueStr);
			m.find();
			// don't check if pattern exist, since it has to
			double value = Double.parseDouble(valueStr.substring(0, m.end()));

			result.add(new PerfDataEntry(hostname, servicedescription, label,
					timestamp, value, state));

			// find next entry, separated by a space
			i = perfdata.indexOf(" ", labelend + 1);
			if (i > -1) {
				++i; // skip that space
			}
		}

		return result;
	}

	/**
	 * Waits for the {@link SpoolFilesWatchDog} to add paths to the internal
	 * queue and initiates the processing of those paths by the
	 * {@link #read(Path)} method. After successful reading (and therefore
	 * storing in the database) a performance data file the corresponding
	 * {@link CheckTimeSeries} instance is informed.
	 */
	@Override
	public void run() {
		Thread thisThread = Thread.currentThread();
		while (reader == thisThread) {
			Path path;
			try {
				path = queue.take();
			} catch (InterruptedException e) {
				if (reader == null) {
					LOGGER.debug("Received shut down interrupt.");
				} else {
					LOGGER.warn("Interrupting.", e);
				}
				return;
			}
			boolean success = read(path);
			if (success) {
				synchronized (checkTimeSeries) {
					checkTimeSeries.notify();
				}
			} // TODO else queue path again?
		}
	}

	/**
	 * Groups all relevant informations of a performance data entry for
	 * simplified handling.
	 */
	private class PerfDataEntry {
		private final String hostname;
		private final String servicedescription;
		private final String perfdatalabel;
		private final int timestamp;
		private final double value;
		private final int state;

		public PerfDataEntry(String hostname, String servicedescription,
				String perfdatalabel, int timestamp, double value, int state) {
			this.hostname = hostname;
			this.servicedescription = servicedescription;
			this.perfdatalabel = perfdatalabel;
			this.timestamp = timestamp;
			this.value = value;
			this.state = state;
		}

		@Override
		public String toString() {
			return String.format("%d [%s - %s] %s: %f", timestamp, hostname,
					servicedescription, perfdatalabel, value);
		}
	}
}

