/*
 *  Copyright (C) 2015  Malte Humann
 *
 *  This file is part of the Time Series Anomaly Detection (TSAD).
 *
 *  TSAD is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TSAD is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with TSAD.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.tzi.imonitor;

import static java.nio.file.LinkOption.NOFOLLOW_LINKS;
import static java.nio.file.StandardWatchEventKinds.ENTRY_CREATE;
import static java.nio.file.StandardWatchEventKinds.OVERFLOW;

import java.io.IOException;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Monitors a given directory for changes and informs the corresponding
 * {@link SpoolFilesReader reader} component about new files through a shared
 * queue.
 * <p>
 * Also provides {@link #start()} and {@link #stop()} methods to safely stop the
 * execution of this thread.
 */
public class SpoolFilesWatchDog implements Runnable {
	/** Reference to the current thread itself to be able to safely stop it. */
	private volatile Thread watchdog;
	/** The directory to be watched for changes. */
	private Path dir;
	/** The watch service actually monitoring the directory. */
	private WatchService watcher;
	private WatchKey key;
	/**
	 * The instance of the spool files reader this instance forwards new files
	 * to.
	 */
	private SpoolFilesReader reader;
	private final static Logger LOGGER = LoggerFactory
			.getLogger(SpoolFilesWatchDog.class);

	/**
	 * Initializes the underlying {@link #watcher watch service} for the given
	 * directory.
	 * 
	 * @param reader
	 *            the spool files reader instance this watch dog forwards new
	 *            files to
	 * @param dir
	 *            the directory to monitor for changes
	 * @throws IOException
	 */
	SpoolFilesWatchDog(SpoolFilesReader reader, Path dir) throws IOException {
		this.dir = dir;
		this.reader = reader;
		watcher = FileSystems.getDefault().newWatchService();
		key = dir.register(watcher, ENTRY_CREATE);
	}

	/** Starts a new thread for the current instance. */
	public void start() {
		watchdog = new Thread(this);
		watchdog.start();
	}

	/** Stops the execution of the current {@link #watchdog thread} safely. */
	public synchronized void stop() {
		LOGGER.debug("Stopping.");
		Thread tmp = watchdog;
		watchdog = null;
		tmp.interrupt();
	}

	/**
	 * Waits for the {@link #watcher} to report any changes and adds the
	 * relevant files to the {@link #reader}'s queue.
	 * <p>
	 * (The reader gets automatically notified about those additions through the
	 * chosen queue implementation.)
	 */
	@Override
	public void run() {
		Thread thisThread = Thread.currentThread();
		// TODO queue (probably unprocessed) files (again)
		while (watchdog == thisThread) {
			WatchKey currentkey;
			try {
				currentkey = watcher.take();
			} catch (InterruptedException e) {
				if (watchdog == null) {
					LOGGER.debug("Received shut down interrupt.");
				} else {
					LOGGER.warn("Interrupting.", e);
				}
				return;
			}

			if (currentkey != key) {
				LOGGER.debug("Untracked Watchkey '" + currentkey
						+ "'; skipping.");
				continue;
			}

			for (WatchEvent<?> event : key.pollEvents()) {
				if (event.kind() == OVERFLOW) {
					continue;
				}

				Path child;
				try {
					child = dir.resolve((Path) event.context());
				} catch (NullPointerException e) {
					LOGGER.debug("Unable to retrieve context for current event.");
					continue;
				} finally {
					if (!key.reset()) {
						LOGGER.debug("Key no longer valid.");
						AnomalyDetection.stop();
						return;
					}
				}

				if (Files.isRegularFile(child, NOFOLLOW_LINKS)) {
					LOGGER.debug(String.format("queueing %s (%s)", child, event
							.kind().name()));
					reader.queueFile(child);
				}
			}
		}
	}
}

